/*  This file is part of qxv,
 *  http://labs.freehackers.org/projects/show/qxv

   Copyright (C) 2010 Thomas Capricelli <orzel@freehackers.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of version 2 of the GNU General Public
   License as published by the Free Software Foundation

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <QKeyEvent>
#include <QAction>
#include <QMenu>
#include <QMenuBar>
#include <QScrollBar>
#include <QDir>
#include <QFileDialog>
#include <QDesktopWidget>
#include <QPrintDialog>
#include <QMainWindow>
#include <QMessageBox>
#include <QPainter>

#include "displaywidget.h"
#include "displaywindow.h"
#include "application.h"
#include "controlwindow.h"

Application::Application(int & argc, char ** argv)
    :QApplication(argc,argv)
{

    qDebug("Considering %d images", argc-1);
    QStringList basenames;
    for(int i=1; i<argc; i++) {
        QFileInfo fi(argv[i]);
        filenames.append(argv[i]);
        basenames.append(fi.fileName());
    }

    // Create Gui stuff
    displayWindow = new DisplayWindow();
    controlWindow = new ControlWindow(this, basenames);
    createActions();
    createMenus();

    displayWidget = new DisplayWidget(displayWindow, *this);
    displayWindow->setCentralWidget(displayWidget);
    displayWindow->setWindowTitle(tr("qxv 0.2"));
    displayWindow->show();

    connect(displayWidget, SIGNAL(emitKey(QKeyEvent*)), this, SLOT(receiveKey(QKeyEvent*)));
    connect(displayWidget, SIGNAL(rmbClicked()), this, SLOT(rmbClicked()));
    connect(displayWindow, SIGNAL(resized(int,int)), this, SLOT(displayWindowResized(int,int)));
#if 0
    // for debugging..

    QWidget *desktop =  qdw->screen();
    qDebug("\tdesktop Geometry: %d %d", desktop->width(), desktop->height());

    int n = qdw->screenCount();
    QRect r;
    qDebug("%d screens", n);
    for(int i=0; i<n; i++) {
        r= qdw->screenGeometry(i);
        qDebug("\tscreenGeometry (%d): %d,%d %dx%d", i, r.x(), r.y(), r.width(), r.height());
        r= qdw->availableGeometry(i);
        qDebug("\tavailableGeometry (%d): %d,%d %dx%d", i, r.x(), r.y(), r.width(), r.height());
    }
    int myscreen = qdw->screenNumber(displayWindow);
    qDebug("myscreen : %d", myscreen);

    qDebug("w/h : %d %d", qdw->width(), qdw->height());
    qDebug("size(): %d %d", qdw->size().width(), qdw->size().height());
    r= qdw->availableGeometry();
    qDebug("availableGeometry (): %d %d", r.width(), r.height());
    qDebug("\n");
#endif
}

Application::~Application(void)
{
    delete displayWindow;
    delete controlWindow;
}

void Application::open()
{
    QString fileName = QFileDialog::getOpenFileName(0, tr("Open File"), QDir::currentPath());
    if (!fileName.isEmpty())
        doLoad(fileName);
}

void Application::doLoad(const QString &filename)
{
    QImage image(filename);
    if (image.isNull()) {
        QMessageBox::information(0, tr("qxv"), tr("Cannot load %1.").arg(filename));
        return;
    }
    printAct->setEnabled(true);

    originalImage = image;
    backToOriginal();
}

void Application::print()
{
    Q_ASSERT(displayWidget->pixmap());
#ifndef QT_NO_PRINTER
    QPrintDialog dialog(&printer);
    if (dialog.exec()) {
        QPainter painter(&printer);
        QRect rect = painter.viewport();
        QSize size = displayWidget->pixmap()->size();
        size.scale(rect.size(), Qt::KeepAspectRatio);
        painter.setViewport(rect.x(), rect.y(), size.width(), size.height());
        painter.setWindow(displayWidget->pixmap()->rect());
        painter.drawPixmap(0, 0, *displayWidget->pixmap());
    }
#endif
}

void Application::about()
{
    QMessageBox::about(0, tr("About qxv"),
            tr("<p><b>qxv</b> is a basic image viewer, with an interface as close as possible to the old and unmaintained <b>xv</b>.</p>") +
            tr("<p><b>qxv</b> is Copyright 2010 Thomas Capricelli <orzel@freehackers.org>.</p>")
            );
}

void Application::createActions()
{
    openAct = new QAction(tr("&Open..."), this);
    openAct->setShortcut(tr("Ctrl+O"));
    connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

    printAct = new QAction(tr("&Print..."), this);
    printAct->setShortcut(tr("Ctrl+P"));
    printAct->setEnabled(false);
    connect(printAct, SIGNAL(triggered()), this, SLOT(print()));

    quitAct = new QAction(tr("E&xit"), this);
    quitAct->setShortcut(tr("Ctrl+Q"));
    connect(quitAct, SIGNAL(triggered()), this, SLOT(quit()));

    aboutAct = new QAction(tr("&About"), this);
    connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));

    // Xv actions
    shrinkAct = new QAction(tr("&shrink"), this);
    shrinkAct->setShortcut(QString(","));
    connect(shrinkAct, SIGNAL(triggered()), this, SLOT(shrink()));

    growAct = new QAction(tr("&grow"), this);
    growAct->setShortcut(QString("."));
    connect(growAct, SIGNAL(triggered()), this, SLOT(grow()));

    bigShrinkAct = new QAction(tr("&bigShrink"), this);
    bigShrinkAct->setShortcut(QString("<"));
    connect(bigShrinkAct, SIGNAL(triggered()), this, SLOT(bigShrink()));

    bigGrowAct = new QAction(tr("&bigGrow"), this);
    bigGrowAct->setShortcut(QString(">"));
    connect(bigGrowAct, SIGNAL(triggered()), this, SLOT(bigGrow()));

    rotateClockwiseAct = new QAction(tr("&rotateClockwise"), this);
    rotateClockwiseAct->setShortcut(QString("t"));
    connect(rotateClockwiseAct, SIGNAL(triggered()), this, SLOT(rotateClockwise()));

    rotateCounterClockwiseAct = new QAction(tr("&rotateCounterClockwise"), this);
    rotateCounterClockwiseAct->setShortcut(QString("T"));
    connect(rotateCounterClockwiseAct, SIGNAL(triggered()), this, SLOT(rotateCounterClockwise()));

    backToOriginalAct = new QAction(tr("&backToOriginal"), this);
    backToOriginalAct->setShortcut(QString(Qt::Key_Return));
    connect(backToOriginalAct, SIGNAL(triggered()), this, SLOT(backToOriginal()));

    flipHorizontalAct = new QAction(tr("&flipHorizontal"), this);
    flipHorizontalAct->setShortcut(QString("h"));
    connect(flipHorizontalAct, SIGNAL(triggered()), this, SLOT(flipHorizontal()));

    flipVerticalAct = new QAction(tr("&flipVertical"), this);
    flipVerticalAct->setShortcut(QString("v"));
    connect(flipVerticalAct, SIGNAL(triggered()), this, SLOT(flipVertical()));

    smoothAct = new QAction(tr("&smooth"), this);
    smoothAct->setShortcut(QString("s"));
    connect(smoothAct, SIGNAL(triggered()), this, SLOT(smooth()));

    cropAct = new QAction(tr("&crop"), this);
    cropAct->setShortcut(QString("c"));
    connect(cropAct, SIGNAL(triggered()), this, SLOT(crop()));

    resizeOriginalAct = new QAction(tr("&Normal size"), this);
    resizeOriginalAct->setShortcut(QString("n"));
    connect(resizeOriginalAct, SIGNAL(triggered()), this, SLOT(resizeOriginal()));

    resizeMaximumAct = new QAction(tr("&Maximum size"), this);
    resizeMaximumAct->setShortcut(QString("m"));
    connect(resizeMaximumAct, SIGNAL(triggered()), this, SLOT(resizeMaximum()));

    resizeMaximumRatioAct = new QAction(tr("&Maximum size, keeping ratio"), this);
    resizeMaximumRatioAct->setShortcut(QString("M"));
    connect(resizeMaximumRatioAct, SIGNAL(triggered()), this, SLOT(resizeMaximumRatio()));

    resizeReAspectAct = new QAction(tr("&Restore aspect ratio from original image"), this);
    resizeReAspectAct->setShortcut(QString("a"));
    connect(resizeReAspectAct, SIGNAL(triggered()), this, SLOT(resizeReAspect()));

    resize4x3Act = new QAction(tr("Resize to &4x3"), this);
    resize4x3Act->setShortcut(QString("4"));
    connect(resize4x3Act, SIGNAL(triggered()), this, SLOT(resize4x3()));
}

void Application::createMenus()
{
    fileMenu = new QMenu(tr("&File"), controlWindow);
    fileMenu->addAction(openAct);
    fileMenu->addAction(printAct);
    fileMenu->addSeparator();
    fileMenu->addAction(quitAct);

    viewMenu = new QMenu(tr("&View"), controlWindow);
    viewMenu->addSeparator();

    helpMenu = new QMenu(tr("&Help"), controlWindow);
    helpMenu->addAction(aboutAct);

    controlWindow->menuBar()->addMenu(fileMenu);
    controlWindow->menuBar()->addMenu(viewMenu);
    controlWindow->menuBar()->addMenu(helpMenu);
}

void Application::receiveKey(QKeyEvent *event)
{
    // qDebug("Application::receiveKey");
    if (event->text()=="q")
        quit();

    if (event->text()==",") {
        shrinkAct->trigger();
        return;
    }
    if (event->text()==".") {
        growAct->trigger();
        return;
    }
    if (event->text()=="<") {
        bigShrinkAct->trigger();
        return;
    }
    if (event->text()==">") {
        bigGrowAct->trigger();
        return;
    }
    if (event->text()=="n") {
        resizeOriginalAct->trigger();
        return;
    }
    if (event->text()=="m") {
        resizeMaximumAct->trigger();
        return;
    }
    if (event->text()=="M") {
        resizeMaximumRatioAct->trigger();
        return;
    }
    if (event->text()=="a") {
        resizeReAspectAct->trigger();
        return;
    }
    if (event->text()=="4") {
        resize4x3Act->trigger();
        return;
    }
    if (event->text()=="t") {
        rotateClockwiseAct->trigger();
        return;
    }
    if (event->text()=="T") {
        rotateCounterClockwiseAct->trigger();
        return;
    }
    if (event->key()==Qt::Key_Return) {
        backToOriginalAct->trigger();
        return;
    }
    if (event->text()=="h") {
        flipHorizontalAct->trigger();
        return;
    }
    if (event->text()=="v") {
        flipVerticalAct->trigger();
        return;
    }
    if (event->text()=="s") {
        smoothAct->trigger();
        return;
    }
    if (event->text()=="c") {
        cropAct->trigger();
        return;
    }
    if (event->key()==Qt::Key_Backspace) {
        previousImage();
        return;
    }
    if (event->key()==Qt::Key_Space) {
        nextImage();
        return;
    }
}

void Application::previousImage(void)
{
    int idx = controlWindow->getFileNumber();
    if (idx<1) return;
    idx-=1;
    controlWindow->selectImage(idx);
    doLoad(filenames[idx]);
}

void Application::nextImage(void)
{
    int idx = controlWindow->getFileNumber();
    if (idx>=filenames.count()-1) return;
    idx+=1;
    controlWindow->selectImage(idx);
    doLoad(filenames[idx]);
}

void Application::rmbClicked(void)
{
    // qDebug("Application::rmbClicked");
    if (controlWindow->isVisible())
        controlWindow->hide();
    else
        controlWindow->show();
}

void Application::displayWindowResized(int width, int height)
{
//    qDebug("Application::displayWindowResized(%d,%d)", width, height);
    displayedImage = currentImage.scaled(width, height, Qt::IgnoreAspectRatio);
    updateImage();
}

void Application::shrink(void)
{
    displayedImage = currentImage.scaled(
            (displayedImage.width()*9)/10,
            (displayedImage.height()*9)/10,
            Qt::IgnoreAspectRatio);
    updateImage();
}

void Application::grow(void)
{
    int w = (displayedImage.width()*11)/10,
        h = (displayedImage.height()*11)/10;
    if (displayedImage.width()==w) w++;
    if (displayedImage.height()==h) h++;
    displayedImage = currentImage.scaled(
            (displayedImage.width()*11)/10,
            (displayedImage.height()*11)/10,
            Qt::IgnoreAspectRatio);
    updateImage();
}

void Application::bigShrink(void)
{
    displayedImage = currentImage.scaled(displayedImage.width()/2, displayedImage.height()/2, Qt::IgnoreAspectRatio);
    updateImage();
}

void Application::bigGrow(void)
{
    displayedImage = currentImage.scaled(displayedImage.width()*2, displayedImage.height()*2, Qt::IgnoreAspectRatio);
    updateImage();
}

void Application::resizeOriginal(void)
{
    displayedImage = currentImage;
    updateImage();
}

void Application::resizeMaximum(void)
{
    // this is tricky : showMaximized() knows the best how to do this, so
    // we use it. This will trigger a resizeEvent(), and
    // Application::displayWindowResized will update the image displayed,
    // and then the window. A superfluous resize will be done by
    // updateImage, but this should be a no-op thanks to qt optimization,
    // hopefully..
    displayWindow->showMaximized();
    return;

    // Could be used, but doesn't know about window frame/decorations
    QRect available = QApplication::desktop()->availableGeometry(displayWindow);
    displayedImage = currentImage.scaled(available.width(), available.height(), Qt::IgnoreAspectRatio);
    // updateImage() will try to update the size without the position,
    // better do it now
    available.setSize(displayedImage.size());
//    displayWindow->setGeometry(available);
//    qDebug("\tsetGeometry: %d,%d %dx%d", available.x(), available.y(), available.width(), available.height());
    updateImage();
}

void Application::resizeMaximumRatio(void)
{
    // available space on screen:
    QRect available = QApplication::desktop()->availableGeometry(displayWindow);
    // available space if we want the window decoration/frame to be visible:
    int horizontalFrameSize = displayWindow->frameGeometry().width() - displayWindow->width();
    int verticalFrameSize = displayWindow->frameGeometry().height() - displayWindow->height();
    // qDebug("Frame/decoration overhead : %d x %d", horizontalFrameSize, verticalFrameSize);
    int width = available.width() - horizontalFrameSize;
    int height = available.height() - verticalFrameSize;

    // resize image
    displayedImage = currentImage.scaled(width, height, Qt::KeepAspectRatio);

    // update window
    updateImage();
}


void Application::resizeReAspect(void)
{
    int width=displayedImage.width(), height=displayedImage.height();
    // it seems to me this would be better, by only growing the appropriate
    // direction, to keep as much information as possible. As does the
    // ReAspect work..
    if (displayedImage.width()*currentImage.height() > displayedImage.height()*currentImage.width()) {
        // not high enough
        height = int(float(width)*float(currentImage.height())/float(currentImage.width()));
    } else {
        // not large enough
        width = int(float(height)*float(currentImage.width())/float(currentImage.height()));
    }
    displayedImage = currentImage.scaled(width, height, Qt::IgnoreAspectRatio);
    updateImage();
}

void Application::resize4x3(void)
{
    int width=displayedImage.width(), height=displayedImage.height();
#if 0 
    // it seems to me this would be better, by only growing the appropriate
    // direction, to keep as much information as possible. As does the
    // ReAspect work..
    if (displayedImage.width()*3 > displayedImage.height()*4) {
        // not high enough
        height = int(float(width)*3./4.);
    } else {
        // not large enough
        width = int(float(height)*4./3.);
    }
#else
    // It seems weird, but the original xv always adjust height according
    // to width..
    height = int(float(width)*3./4.);
#endif
    displayedImage = currentImage.scaled(width, height, Qt::IgnoreAspectRatio);
    updateImage();
}

void Application::rotateClockwise(void)
{
    QMatrix m;
    m.rotate(90);
    displayedImage = displayedImage.transformed(m);
    currentImage = currentImage.transformed(m);
    updateImage();

    // QRect has no transpose() :/
    QSize size = displayedRect.size();
    size.transpose();
    displayedRect.setSize(size);
}

void Application::rotateCounterClockwise(void)
{
    QMatrix m;
    m.rotate(-90);
    displayedImage = displayedImage.transformed(m);
    currentImage = currentImage.transformed(m);

    // QRect has no transpose() :/
    QSize size = displayedRect.size();
    size.transpose();
    displayedRect.setSize(size);
}

void Application::backToOriginal(void)
{
    currentImage = displayedImage = originalImage;
    updateImage();

    displayedRect = originalImage.rect();
}

void Application::flipHorizontal(void)
{
    displayedImage = displayedImage.mirrored(true, false);
    currentImage = currentImage.mirrored(true, false);
    updateImage();
    // no change for displayedRect
}

void Application::flipVertical(void)
{
    displayedImage = displayedImage.mirrored(false, true);
    currentImage = currentImage.mirrored(false, true);
    updateImage();
    // no change for displayedRect
}

void Application::crop(void)
{
    QRect rect = displayWidget->selection();
    if (rect.isNull()) return;
//    qDebug("crop(void)");

    // computing displayedRect is tricky
    // rescaling
    float xfactor = float(displayedRect.width())/float(displayedImage.width());
    float yfactor = float(displayedRect.height())/float(displayedImage.height());
    QRect scaled = QRect(
            rect.x() * xfactor,
            rect.y() * yfactor,
            rect.width() * xfactor,
            rect.height() * yfactor
            );
    displayedRect = scaled.translated(displayedRect.topLeft());

    currentImage = displayedImage = displayedImage.copy(rect);
    updateImage();
}

void Application::smooth(void)
{
    qDebug("smooth(void)");
}

void Application::loadImage()
{
    int idx = controlWindow->getFileNumber();
    Q_ASSERT(idx>=0);
    Q_ASSERT(idx<filenames.count());
    doLoad(filenames[idx]);
}

void Application::updateImage(void)
{
#if 0
    qDebug("displayedImage size : %dx%d", displayedImage.width(), displayedImage.height());
    qDebug("displayWindow min/maximumSize : %dx%d, and %dx%d",
            displayWindow->minimumSize().width(), displayWindow->minimumSize().height(),
            displayWindow->maximumSize().width(), displayWindow->maximumSize().height());
    qDebug("widget min/maximumSize : %dx%d, and %dx%d",
            displayWidget->minimumSize().width(), displayWidget->minimumSize().height(),
            displayWidget->maximumSize().width(), displayWidget->maximumSize().height());
#endif
    displayWindow->resize(displayedImage.size());
    displayWidget->syncPixmap();
}


QColor Application::getPixel(QPoint &pos) const
{
    // We need to find out the position in the original image
    float xfactor = float(displayedRect.width())/float(displayedImage.width());
    float yfactor = float(displayedRect.height())/float(displayedImage.height());
    int x = displayedRect.x() + xfactor * pos.x();
    int y = displayedRect.y() + yfactor * pos.y();
    // we can't use originalImage.rect().contains(pos) as we need to clip,
    // not just 'know'
    x = qMax(0, x);
    x = qMin(originalImage.rect().right(), x);
    y = qMax(0, y);
    y = qMin(originalImage.rect().bottom(), y);

    // update pos
    pos = QPoint(x,y);

    return QColor(originalImage.pixel(pos));
}

//    qDebug("currentimage size : %dx%d", displayedImage.width(), displayedImage.height());
