/*  This file is part of qxv,
 *  http://labs.freehackers.org/projects/show/qxv

   Copyright (C) 2010 Thomas Capricelli <orzel@freehackers.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of version 2 of the GNU General Public
   License as published by the Free Software Foundation

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/* Qt */
#include <QMouseEvent>
#include <QPainter>

/* Project */
#include "rubberband.h"
#include "displaywidget.h"
#include "application.h"

DisplayWidget::DisplayWidget(QWidget *parent, const Application &_application, Qt::WindowFlags f)
    : QLabel(parent, f)
    , displayInformationDisplayedOnTop(true)
    , application(_application)
{
    setFocusPolicy(Qt::StrongFocus);
    rubberBand = new RubberBand(this);
    dragState=DragNone;
}

void DisplayWidget::keyPressEvent(QKeyEvent *event)
{
    emit emitKey(event);
    // else.. call base class implementation
    //QLabel::keyPressEvent(event);
}

void DisplayWidget::syncPixmap()
{
    rubberBand->hide();
    displayedPixmap = QPixmap::fromImage(application.displayedImage);
    setPixmap(displayedPixmap);
}

void DisplayWidget::mousePressEvent(QMouseEvent *event)
{
    // qDebug("DisplayWidget::mousePressEvent");
    if (event->button()== Qt::RightButton) {
        emit rmbClicked();
        return;
    }
    switch(event->button()) {
        case Qt::RightButton:
            emit rmbClicked();
            break;
        case Qt::MiddleButton:
            dragState = DragPixelInformation;
            origin = event->pos();
            originOriginal = origin;
            application.getPixel(originOriginal); // compute origin in original image reference
            displayInformationDisplayedOnTop = origin.y() >= height()/2;
            updatePixelInformation(event->pos());
            break;
        case Qt::LeftButton:
            dragState = DragSelection;
            origin = event->pos();
            rubberBand->setGeometry(QRect(origin, QSize()));
            rubberBand->show();
            break;
        default: // ???
            break;
    }

}

void DisplayWidget::mouseMoveEvent(QMouseEvent *event)
{
    switch (dragState) {
        case DragNone:
            // nothing
            break;
        case DragSelection:
            rubberBand->setGeometry(QRect(origin, event->pos()).normalized() & parentWidget()->rect());
            break;
        case DragPixelInformation:
            updatePixelInformation(event->pos());
            break;
    }
}

void DisplayWidget::mouseReleaseEvent(QMouseEvent *)
{
    switch (dragState) {
        case DragNone:
            // nothing
            break;
        case DragSelection:
            // rubberBand->hide();
            dragState=DragNone;
            break;
        case DragPixelInformation:
            syncPixmap(); // erase text
            break;
    }
}

QRect DisplayWidget::selection(void)
{
    if (!rubberBand->isVisible())
        return QRect();
    QRect rect = rubberBand->geometry();
    return rect;
}

void DisplayWidget::updatePixelInformation(QPoint pos)
{
    // qDebug("colorpicker at %d %d", pos.x(), pos.y());

    // Get color and update 'pos'
    QColor color = application.getPixel(pos);

    // Construct info string
    // for whatever reason, xv displays them on a scale of [0..100]
    int hue = qMax(color.hsvHue(), 0);
    int saturation = color.hsvSaturationF()*100;
    int value = color.valueF()*100.;
    QString info=QString(" %1, %2 = %3,%4,%5 %6 (%7 %8 %9 HSV) [%10, %11]").arg(pos.x(),4).arg(pos.y(),4)
        .arg(color.red(),3).arg(color.green(),3).arg(color.blue(),3)
        .arg(color.name(),7)
        .arg(hue,3).arg(saturation,3).arg(value,3)
        .arg(pos.x()-originOriginal.x(),4)
        .arg(pos.y()-originOriginal.y(),4);
    // qDebug() << info;

    // display it
    // The use of a local copy is to prevent rubbish to remain when we
    // display a string shorter than the previous one. This is not really
    // needed here as we use (on purpose) a fixed-width font and a
    // fixed-width string, but just in case...
    QPixmap localCopy(displayedPixmap);
    QPainter p(&localCopy);
    QFont f("Courier", 11);
    f.setLetterSpacing(QFont::AbsoluteSpacing, -2);
    p.setFont(f);
    // boundingRect() ignores a leading space in info.
    int adjustForBugInQtFontMetrics = 2*info.count();
    QRect r = p.fontMetrics().boundingRect(info).adjusted(0,0,7-adjustForBugInQtFontMetrics,0);
    if (displayInformationDisplayedOnTop)
        r.moveTopLeft(QPoint(0,0));
    else
        r.moveTopLeft(QPoint(0,displayedPixmap.height()-r.height()));
    p.fillRect(r, Qt::black);
    p.setPen(Qt::white);
    p.drawText(0,r.y()+p.fontMetrics().ascent(), info);
    p.end();
    setPixmap(localCopy);
}

